// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//


package de.uka.ilkd.key.java.recoderext;

import recoder.*;
import recoder.java.*;
import recoder.java.statement.*;
import recoder.java.expression.*;
import recoder.java.reference.*;
import recoder.list.*;

/**
 *  A shortcut-statement for a method body.
 */

public class MethodBodyStatement extends JavaStatement implements
    TypeReferenceContainer, ExpressionContainer {


    /**
     * the ast parent
     */
    private StatementContainer astParent;

    protected TypeReference bodySource;
    protected Expression resultVar;
    protected MethodReference methRef;
    
    /**
     *      Construct a method body shortcut
     *      @param bodySource exact class where the body is declared
     *      @param resultVar the Expression naming the variable to which
     *      the result of the mthod is assigned
     *      @param methRef MethodReference that represents the call
     */
    
    public MethodBodyStatement(TypeReference bodySource, 
			       Expression resultVar,
			       MethodReference methRef) {
        setBodySource(bodySource);
	this.resultVar  = resultVar;
        setMethodReference(methRef);
        makeParentRoleValid();
    }

    /**
     *      Construct a method body shortcut
     *      @param proto a MethodBodyStatement
     */
    
    protected MethodBodyStatement(MethodBodyStatement proto) {
        super(proto);
        if (proto.bodySource != null) {
            bodySource = (TypeReference) proto.bodySource.deepClone();
        }
        if (proto.resultVar != null) {
            resultVar = (Expression) proto.resultVar.deepClone();
        }
        if (proto.methRef != null) {
            methRef = (MethodReference) proto.methRef.deepClone();
        }
        makeParentRoleValid();
    }

    /**
     *	Set the exact Class the denoted method body is from.
     */
    
    public void setBodySource(TypeReference bodySource) {
	this.bodySource = bodySource;
    }

    /**
     *	Get the exact Class the denoted method body is from.
     *  @returns the TypeReference
     */
    
    public TypeReference getBodySource() {
	return bodySource;
    }

    /**
     *      Set the result variable.
     *      @param the VariableReference
     */
    
    public void setResultVariable(Expression resultVar) {
	this.resultVar = resultVar;
    }

    /**
     *      Get the result variable.
     *      @return the VariableReference
     */
    
    public Expression getResultVariable() {
	return resultVar;
    }


    /**
     *      Set the MethodReference that caused this call.
     */
    
    public void setMethodReference(MethodReference methRef) {
        this.methRef = methRef;
    }

    /**
     *      Get the MethodReference that caused this call.
     *      @return the MethodReference
     */
    public MethodReference getMethodReference() {
        return methRef;
    }

    public NonTerminalProgramElement getASTParent() {
	return astParent;
    }

    public StatementContainer getStatementContainer() {
	return astParent;
    }

    public void setStatementContainer(StatementContainer parent) {
	astParent = parent;
    }

    /**
     *      Finds the source element that occurs first in the source.
     *      @return the last source element in the syntactical representation of
     *      this element, may be equals to this element.
     */
    
    public SourceElement getFirstElement() {
        return getChildAt(0).getFirstElement();
    }

    /**
     *      Finds the source element that occurs last in the source.
     *      @return the last source element in the syntactical representation of
     *      this element, may be equals to this element.
     */
    
    public SourceElement getLastElement() {
        return getChildAt(getChildCount() - 1).getLastElement();
    }

    /**
     *      Returns the number of children of this node.
     *      @return an int giving the number of children of this node
     */
    
    public int getChildCount() {
        int result = 0;
        if (bodySource != null) result++;
        if (resultVar != null) result++;
        if (methRef != null) result++;
        return result;
    }

    /**
     *      Returns the child at the specified index in this node's "virtual"
     *      child array
     *      @param index an index into this node's "virtual" child array
     *      @return the program element at the given position
     *      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
     *                 of bounds
     */
    
    public ProgramElement getChildAt(int index) {
        if (bodySource != null) {
            if (index == 0) return bodySource;
            index--;
        }
        if (resultVar != null) {
            if (index == 0) return resultVar;
            index--;
        }
        if (methRef != null) {
            if (index == 0) return methRef;
            index--;
        }
        throw new ArrayIndexOutOfBoundsException();
    }

    /**
     *      Get the number of type references in this container.
     *      @return the number of statements.
     */

    public int getTypeReferenceCount() {
        return (bodySource != null) ? 1 : 0;
    }

    /**
     *       Return the type reference at the specified index in this node's
     *       "virtual" statement array.
     *       @param index an index for a type reference.
     *       @return the type reference with the given index.
     *       @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
     *       of bounds.
     */

    public TypeReference getTypeReferenceAt(int index) {
        if (bodySource != null && index == 0) {
            return bodySource;
        }
        throw new ArrayIndexOutOfBoundsException();
    }
    

    /**
     *      Get the number of expressions in this container.
     *      @return the number of expressions.
     */

    public int getExpressionCount() {
        int result = 0;
        if (methRef != null) result++;
        if (resultVar != null) result++;
        return result;
    }

    /**
     *       Return the expression at the specified index in this node's
     *       "virtual" expression array.
     *       @param index an index for a expression.
     *       @return the expression with the given index.
     *       @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
     *       of bounds.
     */

    public Expression getExpressionAt(int index) {
	if (methRef != null) {
	    if (index == 0) return methRef;
	    index--;
	} 
	if (resultVar != null) {
	    if (index == 0) return resultVar;
	}
        throw new ArrayIndexOutOfBoundsException();
    }

    /**
     * Replace a single child in the current node.
     * The child to replace is matched by identity and hence must be known
     * exactly. The replacement element can be null - in that case, the child
     * is effectively removed.
     * The parent role of the new child is validated, while the
     * parent link of the replaced child is left untouched.
     * @param p the old child.
     * @param p the new child.
     * @return true if a replacement has occured, false otherwise.
     * @exception ClassCastException if the new child cannot take over
     *            the role of the old one.
     */

    public boolean replaceChild(ProgramElement p, ProgramElement q) {
        if (p == null) throw new NullPointerException();
        if (bodySource == p) {
            TypeReference r = (TypeReference)q;
            bodySource = r;
            if (r != null) {
                r.setParent(this);
            }
            return true;
        }
        else if (resultVar == p)
        {
            Expression r = (Expression)q;
            resultVar = r;
            if (r != null) {
                r.setExpressionContainer(this);
            }
            return true;
        }
        else if (methRef == p) {
            MethodReference r = (MethodReference)q;
            methRef = r;
            if (r != null) {
                r.setExpressionContainer(this);
            }
            return true;
        }
        return false;
    }
    
    /**
     *      Ensures that each child has "this" as syntactical parent.
     */

    public void makeParentRoleValid() {
        super.makeParentRoleValid();

        if (bodySource != null) {
            bodySource.setParent(this);
        }

        if (resultVar != null) {
            resultVar.setExpressionContainer(this);
        }

        if (methRef != null) {
            methRef.setExpressionContainer(this);
        }
    }

    public int getChildPositionCode(ProgramElement child) {
        // role 0: bodySource
	// role 1: resultVar
        // role 2: methRef
        if (bodySource == child) {
            return 0;
        }
	if (resultVar == child) {
	    return 1;
        }
        if (methRef == child) {
            return 2;
        }
        return -1;
    }

    //don't think we need it
    public void accept(SourceVisitor v) {
    }
    
    /**
     * Deep clone.
     * @return the object
     */
    public Object deepClone() {
	return new MethodBodyStatement(this);
    }

}
