// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.proof.mgt;

import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.java.JavaProgramElement;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.Expression;
import de.uka.ilkd.key.java.ArrayOfExpression;
import de.uka.ilkd.key.java.TypeConverter;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.expression.operator.*;
import de.uka.ilkd.key.java.StatementBlock;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.JavaBlock;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.util.*;
import de.uka.ilkd.key.casetool.*;
import de.uka.ilkd.key.java.Statement;
import de.uka.ilkd.key.java.reference.ConstructorReference;
import de.uka.ilkd.key.java.expression.Assignment;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.proof.init.*;
import java.util.*;
import de.uka.ilkd.key.java.statement.MethodBodyStatement;
import de.uka.ilkd.key.logic.sort.ProgramSVSort;
import de.uka.ilkd.key.java.declaration.MethodDeclaration;

public class MethodLemmaFactory {
    
    private ReprModelMethod repr;
    private ModifiesParserHelper modPH = null;
    private InitConfig initConfig;
    private String pre;
    private String post;
    private String suffix;

    private static final ArrayOfQuantifiableVariable emptyvars 
	= new ArrayOfQuantifiableVariable();

    public MethodLemmaFactory(ReprModelMethod repr, InitConfig initConfig,
			      String pre, String post, String suffix) {
	this.repr = repr;
	this.initConfig = initConfig;
	this.pre = pre;
	this.post = post;
	this.suffix = suffix;
    }

    public ReprModelMethod getModelMethod() {
	return repr;
    }


    //-------- below the stuff to create a taclet that inserts this method spec 

    private OCLPostPO postPO = null;
    private ArrayOfQuantifiableVariable proofQuantVars = null;
    private SortedSchemaVariable result = null;
    private Term resultTerm = null;
    private HashMap schemaVariablesMap;
    private ProgramElement proofStatement;
    private TypeConverter typeConv;
    private TermSVReplacer tSVR;
    private Term proofPrefix;
    private Term proofPrecondition;
    private Term proofPurePrefix;
    private Term proofSuffix;
    private Term pi2Upd;
    private Term[] modSUTArr; 
    private QuantifiableVariable[] logicQuantVarsArr;
    private Set modElmsAsTerms;
    private ArrayOfQuantifiableVariable logicQuantVars;
    private Term psiUpd = null;
    private Term addFormula = null;
    private Term schVarPsiTerm;
    private Statement transformedStatement;
    private Term findTerm;
    private Namespace ns;
    private Map lV2SV;


    public Taclet createTaclet() {

	if (repr.getMyModifClause()==null && !repr.isQuery()) {
	    return null;
	}
	
	// gets the OCL-DL translated proof which is also already transformed
	// by the method described by Sebastian's Studienarbeit so it can be
	// used in the way his formula construction needs it
	Term proof = createOCLtoDLTransformedProof();
	
	// there is an error in the pre or postcondition and therefore no
	// taclet can be created by the given rule
	if (proof == null) {
	    return null;
	}
	
	// puts the eliminated quantifiers into proofQuantVars which
	// is an ArrayOfQuantifiableVariable
	Term nonQuantifiedProof = 
	    eliminateQuantificationRetainsQuantifyingVariables(proof);
	
	// partition the now unquantified proof
	proofPrefix = nonQuantifiedProof.sub(0);
	proofPrecondition = proofPrefix.sub(0);//PRE,i.e.Phi-Pre before map
	proofPurePrefix = proofPrefix.sub(1);  //Pi1
	JavaBlock proofJavaBlock = nonQuantifiedProof.sub(1).javaBlock();
	proofSuffix = nonQuantifiedProof.sub(1).sub(0); //Pi2
	
	// by side effect this gives us:
	// - result
	// - resultTerm
	// - schemaVariablesMap
	// - transformedStatement
	// - typeConv
	createMethodBodyStatementAndGetParameterToSVMap(proofJavaBlock);
	
	// the given proofStatement is of wrong type and therefore no taclet
	// can be createdby the given rule
	if (transformedStatement == null) {
	    return null;
	}
	
	// replaces arguments by schema variables inside terms using
	// schemaVariablesMap in TermSVReplacer
	// changes to the following:
	// - proofPrefix
	// - proofPrecondition
	// - proofPurePrefix
	// - proofSuffix
	// creates:
	// - tSVR -- which is the term replacer
	replaceTerms();
	
	// create the namespace made up of the keys from the schema
	// variable mapping
	ns = new Namespace();
	Set keys = schemaVariablesMap.keySet();
	Iterator keysIt = keys.iterator();
	while (keysIt.hasNext()) {
	    ns.add((Named) keysIt.next());
	}
	
	// finds all modifies elements and returns a Set
	// filled by those already converted to terms and the arguments
	// having been replaced by schemavariables
	modElmsAsTerms = getModifiesElementsAsTerms();
	
	// there is no modifies set given and thus no taclet can be created
	// by the given rule
	if (modElmsAsTerms == null) {
	    return null;
	}
	
	// fills the following:
	// - pi2Upd
	// - modSUTArr
	// - logicQuantVarsArr
	createPi2UpdatedAndSaveLogicQuantifierVars();
	
	// fills the following:
	// - logicQuantVars
	// - psiUpd
	createPsiUpdated();
	
	// first add clause
	Term addPRE = proofPrecondition;
	// second add clause
	addFormula = createAddFormula();

	// creates the find clause
	findTerm = createFindTerm();

	SuccTacletBuilder stb = new SuccTacletBuilder();
	
	// put the find part into the taclet builder
        stb = stb.setFind(findTerm);	
	
	Iterator it = lV2SV.values().iterator();
	while (it.hasNext()) {
	    stb.addVarsNotFreeIn((SchemaVariable)it.next(),
				 (SchemaVariable)schVarPsiTerm.sub(0).op());
	}
	
	// add one add part to the taclet builder
	stb.addTacletGoalTemplate(createTGTfromTerm(addPRE, Boolean.TRUE));
	
	// add the other add part to the taclet builder
	stb.addTacletGoalTemplate(createTGTfromTerm(addFormula, Boolean.FALSE));
	
	// getting the name for the taclet
	if (transformedStatement instanceof MethodBodyStatement) {
	    String stbName = ((MethodBodyStatement) transformedStatement)
		.getProgramMethod().name().toString();
	    stbName = ((MethodBodyStatement) transformedStatement)
		.getBodySource().getName() + "_" + stbName;
	    
	    // display name does not need to be unique thus for better user
	    // readability we only use "classname_methodname" for a method call
	    stb.setDisplayName(stbName+"_"+suffix);
	    
	    MethodDeclaration meth = ((MethodBodyStatement) 
				      transformedStatement)
		.getProgramMethod().getMethodDeclaration();
	    int i = 0;
	    while (i < meth.getParameterDeclarationCount()) {
		stbName = stbName + "_" + meth.getParameterDeclarationAt(i)
		    .getVariableSpecification().getType().getName();
		i = i + 1;
	    }
	    stb.setName(new Name(stbName+"_"+suffix));
	}
	else if (transformedStatement instanceof Assignment) { 
	    String stbName = ((New) ((Assignment) transformedStatement)
			      .getChildAt(1)).getTypeReference().getName();
	    
	    // display name does not need to be unique thus for better user
	    // readability we only use the classname for a constructor
	    stb.setDisplayName(stbName+"_"+suffix);
	    
	    // using proofStatement because there are the real argument types 
	    // whereas in transformedStatement there are only ProgramSV's
	    // after putting in the schema variables 
	    ArrayOfExpression aOEarg = ((New) ((Assignment) proofStatement)
					.getChildAt(1)).getArguments();
	    int i = 0;
	    while (i < aOEarg.size()) {
		stbName = stbName + "_" + 
		    ((ProgramVariable) aOEarg.getExpression(i))
		    .getKeYJavaType().getName();
		i = i + 1;
	    }
	    stb.setName(new Name(stbName+"_"+suffix));
	}
	else {
	    stb.setName(new Name("ERROR has happened! Don't use this!"));
	}
	
	stb.addHeuristic((Heuristic)initConfig.heurNS().lookup
			 (new Name("use_method_specs")));

	SuccTaclet succTac = stb.getSuccTaclet();
	return succTac;
		
    }


    private Term createOCLtoDLTransformedProof() {
	// get the OCL-DL translated Term including pre- and postconditions
	postPO = new OCLPostPO(repr.getName(), repr, pre,
			       post, repr.getActXmifile());
	postPO.setConfig(initConfig);
	NamespaceSet nss = initConfig.namespaces();
	Namespace oldProgVarsNS = nss.programVariables();
	nss.setProgramVariables(new Namespace());
	try{
	    postPO.read(ModStrategy.NO_VARS);
	}
	// in case there is an exception we couldn't parse the given pre- or
	// postcondition and thus can't create a rule, this gets taken care of 
	// just after the call to this method in the main part
	catch (de.uka.ilkd.key.ocl.ocldltranslator.OCLTranslationException e) {
	    System.out.println("Warning: Syntax error in OCL specification of "
			       +repr.getName());
	    return null;
	} finally {
	    nss.setProgramVariables(oldProgVarsNS);
	}
	Term dlProof = postPO.getPOTerm();

	// transform it to the form which is needed for the taclet generation 
	// according to Sebastian's Studienarbeit
	Term proof = MethodSpecTransformation.getTransformedHoare(dlProof);
	
	return proof;
    }
    

    private Term eliminateQuantificationRetainsQuantifyingVariables(Term proof){
	Term nonQuantifiedProof = proof;
	ArrayOfQuantifiableVariable proofQuantVarsHelpArray = null;
	Vector proofQuantVarsVector = new Vector();
	while (nonQuantifiedProof.op() instanceof Quantifier){
	    proofQuantVarsHelpArray = nonQuantifiedProof.varsBoundHere(0);
	    proofQuantVarsVector.add(proofQuantVarsHelpArray.
				     getQuantifiableVariable(0));
	    nonQuantifiedProof = nonQuantifiedProof.sub(0);
	}
	
	Iterator proofQuantVarsVectorIterator = proofQuantVarsVector.iterator();
	QuantifiableVariable[] proofQuantVarsArray = 
	    new QuantifiableVariable[proofQuantVarsVector.size()];
	int count = 0;
	while (proofQuantVarsVectorIterator.hasNext()) {
	    proofQuantVarsArray[count] = 
		(QuantifiableVariable) proofQuantVarsVectorIterator.next();
	    count = count + 1;
	}
	proofQuantVars = new ArrayOfQuantifiableVariable(proofQuantVarsArray);
	
	return nonQuantifiedProof;
    }

    private void createMethodBodyStatementAndGetParameterToSVMap
	(JavaBlock proofJavaBlock) {
	// create methodreference from "ProgramElement" proofStatement and
	// replace parameters by schema variables
	// get the HashMap from parameters(including result) to schemavariables
	StatementBlock proofStatementBlock = 
	    (StatementBlock) proofJavaBlock.program();
	proofStatement = proofStatementBlock.getChildAt(0);
	schemaVariablesMap = new HashMap();
	
	MethodBodyStatement methodBS = null;
	ProgramElement v = null;
	ArrayOfExpression argumentsAoE = null;

	if (proofStatement instanceof Assignment){
	    if ( ((Assignment) proofStatement).getChildAt(1) instanceof ConstructorReference) {
		v = ((Assignment) proofStatement).getChildAt(0);
		argumentsAoE = ((New) ((Assignment) proofStatement).getChildAt(1)).getArguments();
	    }
	    else {
		//ERROR!!!
		System.out.println(" ERROR !!! in createMethodBodyStatementAnd...(...) ");
		transformedStatement = null;
		return;
	    }
	}
	else if (proofStatement instanceof MethodBodyStatement){
	    methodBS = (MethodBodyStatement) proofStatement;
	    v = methodBS.getResultVariable();
	    argumentsAoE = methodBS.getArguments();
	}
	else {
	    //ERROR!!!
	    System.out.println(" ERROR !!! in createMethodBodyStatementAnd...(...) ");
	    transformedStatement = null;
	    return;
	}
	
	
	if (v == null) {
	    result = null;
	    resultTerm = null;
	}
	else {
	    result 
		= (SortedSchemaVariable) SchemaVariableFactory.createProgramSV
		(new ProgramElementName("#result"),
		 ProgramSVSort.VARIABLE, false);
	    schemaVariablesMap.put(v, result);
	    resultTerm = TermFactory.DEFAULT.createVariableTerm(result);
	}
	
	
	Expression[] argumentsArr = new Expression[argumentsAoE.size()];
	for (int i=0; i < argumentsAoE.size(); i++){
	    
	    SortedSchemaVariable mapParameter = (SortedSchemaVariable)
		SchemaVariableFactory
		.createProgramSV(new ProgramElementName
				 ("#parameter_"
				  +((ProgramVariable) 
				    argumentsAoE.getExpression(i)).name()),
				 ProgramSVSort.SIMPLEEXPRESSION, false);
	    
	    argumentsArr[i] = mapParameter;
	    schemaVariablesMap.put(argumentsAoE.getExpression(i),
				   mapParameter);
	}	
	SchemaVariable mapObject = null;
	if (!methodBS.getProgramMethod().isStatic()) {
	    mapObject = SchemaVariableFactory.
		createProgramSV(new ProgramElementName("#prefix"), 
				ProgramSVSort.EXPRESSION, false);
	}
	
	Services services = initConfig.getServices();
	typeConv = services.getTypeConverter();
	
	
	argumentsAoE = new ArrayOfExpression(argumentsArr);
	
	if (proofStatement instanceof MethodBodyStatement) {
	    if (!methodBS.getProgramMethod().isStatic()) {
		schemaVariablesMap.put(methodBS.getDesignatedContext(), 
				       mapObject);
	    }
	    
	    
	    transformedStatement = new MethodBodyStatement
		((ProgramMethod) methodBS.getProgramMethod(),
		 (de.uka.ilkd.key.java.reference.ReferencePrefix) mapObject,
		 (IProgramVariable) result,
		 (ArrayOfExpression) argumentsAoE);
	}
	else { //proofStatement is an instance of an assignment to a
	       //constructor
	    transformedStatement = new CopyAssignment
		(result, 
		 new New
		 (argumentsArr, 
		  ((New) ((Assignment) proofStatement).getChildAt(1))
		  .getTypeReference(), null));
	}
    }

    private void replaceTerms() {
	tSVR = new TermSVReplacer(schemaVariablesMap);
	proofPrefix = tSVR.inst(proofPrefix);
	proofPrecondition = tSVR.inst(proofPrecondition); // PRE
	proofPurePrefix = tSVR.inst(proofPurePrefix); // pi1
	proofSuffix = tSVR.inst(proofSuffix); // pi2
    }

    private Set getModifiesElementsAsTerms() {
	modPH = new ModifiesParserHelper(repr, initConfig);
	try {
	    modPH.parseModifiesClauseToHashSet(ns);
	}
	// in case there is an exception we couldn't parse the given modifies
	// clause and thus can't create a rule, this gets taken care of just
	// after the call to this method in the main part
	catch (Exception e) { 
	    return null;
	}
	Set modElms = modPH.getModifiesElementsHashSet(); 
	//"ProgramElement" members
	
	// in case there is no modifies set return null
	if (modElms == null) {
	    return null;
	}
	                                            
	modElmsAsTerms = new HashSet();
	Iterator modElmsIt = modElms.iterator();
	Term t = null;
	while (modElmsIt.hasNext()) {
	    t = (Term) typeConv.convertToLogicElement
		((ProgramElement) modElmsIt.next());
	    modElmsAsTerms.add(tSVR.inst(t)); // adds the "m_i' " instead of the
	                                      //"m_i" because those are required
	} 
	return modElmsAsTerms;
    }


    private void createPi2UpdatedAndSaveLogicQuantifierVars() {
	Iterator modElmsAsTermsIt = modElmsAsTerms.iterator();
	int counter = 0;
	
	logicQuantVarsArr = null;
	int size;
	if (resultTerm == null){
	    size = modElmsAsTerms.size();	    
	} else {
	    size = modElmsAsTerms.size()+1;	    
	}
	logicQuantVarsArr = new QuantifiableVariable[size];
	modSUTArr = new Term[2*size+1];//modifies Simultaneous Update Term Array
	while (counter < modElmsAsTerms.size()) {
	    Term t = (Term) modElmsAsTermsIt.next();
	    LogicVariable newVar = 
		new LogicVariable(new Name("c_"+ ((AttributeOp)t.op()).attribute().toString()), 
				  t.sort());
	    Term a = TermFactory.DEFAULT.createVariableTerm(newVar);
	    logicQuantVarsArr[counter] = (QuantifiableVariable) newVar;
	    modSUTArr[2*counter] = t;
	    modSUTArr[2*counter+1] = a;
	    counter = counter + 1;
	}
	if (resultTerm !=null) {
	    Sort resultSort = null;
	    if (proofStatement instanceof MethodBodyStatement) {
		resultSort = ((MethodBodyStatement) proofStatement)
		    .getProgramMethod().getKeYJavaType().getSort();
	    }
	    else { // proofStatement is an  Assignment to constructor
		resultSort = ((ProgramVariable) 
			      ((Assignment) proofStatement).getChildAt(0))
		    .getKeYJavaType().getSort();
	    }
	    LogicVariable newVar = new LogicVariable(new Name("_c_"+"result"),
	    					     resultSort);
	    Term a = TermFactory.DEFAULT.createVariableTerm(newVar);

	    logicQuantVarsArr[modElmsAsTerms.size()] = newVar;
	    modSUTArr[2*modElmsAsTerms.size()] = resultTerm;
	    modSUTArr[2*modElmsAsTerms.size()+1] = a;
	    modSUTArr[2*modElmsAsTerms.size()+2] = schVarPsiTerm;	    
	}
	if (size == 0) {
	    pi2Upd = proofSuffix;
	} else {
	    modSUTArr[modSUTArr.length-1] = proofSuffix;
	    pi2Upd = TermFactory.DEFAULT.createUpdateTerm(modSUTArr);
	}
    }
    
    private void createPsiUpdated() {
	SchemaVariable schVarPsi = 
	    SchemaVariableFactory.createFormulaSV
	    (new ProgramElementName("#psi"), false);
	schVarPsiTerm = TermFactory.DEFAULT.createVariableTerm(schVarPsi);

	schVarPsiTerm = TermFactory.DEFAULT.createDiamondTerm
	    (JavaBlock.createJavaBlock(new ContextTerm(new ExtList())), 
	     schVarPsiTerm);
	
	int counter;
	if (resultTerm == null && modElmsAsTerms.size() == 0) {
	    psiUpd = schVarPsiTerm;
	}
	else {
	    Term[] modSUTrArr = null; 
	    //= Modifies Simultaneous Update Term result	
	    modSUTrArr = new Term[modSUTArr.length];
	    for (int i=0; i<modSUTArr.length; i++) {
		modSUTrArr[i]=modSUTArr[i];
	    }
	    modSUTrArr[modSUTrArr.length-1] = schVarPsiTerm;
	    psiUpd = TermFactory.DEFAULT.createUpdateTerm(modSUTrArr);	
	}
	logicQuantVars = new ArrayOfQuantifiableVariable(logicQuantVarsArr);
    }


    private static SortedSchemaVariable getSV(QuantifiableVariable lv, 
					Map logic2Schema) {
	if (logic2Schema.containsKey(lv)) {
	    return (SortedSchemaVariable)logic2Schema.get(lv);
	} else {
	    SortedSchemaVariable sv 
		= (SortedSchemaVariable)
		SchemaVariableFactory.createVariableSV(lv.name(), lv.sort(), 
						       false);
	    logic2Schema.put(lv, sv);
	    return sv;
	}
    }

    private static Term bindLogicVarsAsSchemaVars(Term t, Map logic2Schema) {
	Term[] subs=new Term[t.arity()];
	boolean changed = false;
	ArrayOfQuantifiableVariable[] a
	    = new ArrayOfQuantifiableVariable[t.arity()];
	for (int i=0; i<subs.length; i++) {
	    subs[i]=bindLogicVarsAsSchemaVars(t.sub(i), logic2Schema);
	    if (subs[i]!=t.sub(i)) changed=true;
	    a[i]=emptyvars;
	    QuantifiableVariable[] qs = null;
	    if (t.varsBoundHere(i).size()>0) {
		qs = new QuantifiableVariable[t.varsBoundHere(i).size()];
	    }
	    for (int j=0; j<t.varsBoundHere(i).size(); j++) {
		qs[j]=getSV(t.varsBoundHere(i).getQuantifiableVariable(j),
			    logic2Schema);
		changed=true;
	    }
	    if (t.varsBoundHere(i).size()>0) {
		a[i]=new ArrayOfQuantifiableVariable(qs);		    
	    }
	}
	if (t.op() instanceof LogicVariable) {
	    return TermFactory.DEFAULT.createVariableTerm
		(getSV((LogicVariable)t.op(), logic2Schema));
	} else if (changed){
	    return TermFactory.DEFAULT.createTerm(t.op(), subs, a, 
						  t.javaBlock(), null);
	} else {
	    return t;
	}
    }

    private static ArrayOfQuantifiableVariable getSVs
	(ArrayOfQuantifiableVariable qvs, Map m) {

	QuantifiableVariable[] a = new QuantifiableVariable[qvs.size()];
	for (int i=0; i<qvs.size(); i++) {
	    a[i]=(QuantifiableVariable)m.get(qvs.getQuantifiableVariable(i));
	}
	return new ArrayOfQuantifiableVariable(a);
    }

    private Term createAddFormula() {
	Term pi1 = proofPurePrefix;
	
	TermFactory tF = TermFactory.DEFAULT;
	addFormula = tF.createJunctorTermAndSimplify(Op.AND, pi1, pi2Upd);
	
	lV2SV = new HashMap();
	addFormula = bindLogicVarsAsSchemaVars(addFormula, lV2SV);
	psiUpd = bindLogicVarsAsSchemaVars(psiUpd, lV2SV);

	if (proofQuantVars.size() != 0) {
	    addFormula = tF.createQuantifierTerm
		(Op.EX, getSVs(proofQuantVars, lV2SV), addFormula);
	}
	addFormula = tF.createJunctorTermAndSimplify(Op.IMP, 
						     addFormula, 
						     psiUpd);
	if (logicQuantVars.size() != 0) {
	    addFormula = tF.createQuantifierTerm
		(Op.ALL, getSVs(logicQuantVars, lV2SV), addFormula);	    
	}	
	return addFormula;
    }
    

    private Term createFindTerm() {
	ExtList children = new ExtList();
	children.add(transformedStatement);
	ContextTerm ct = new ContextTerm(children);
	JavaBlock mBSJavaBlock = 
	    JavaBlock.createJavaBlock((JavaProgramElement) ct);
	Term fT = TermFactory.DEFAULT.createDiamondTerm(mBSJavaBlock, 
							schVarPsiTerm.sub(0));
	return fT;
    }

    private TacletGoalTemplate createTGTfromTerm(Term add, Boolean precond) {
	Semisequent semSeq = Semisequent.EMPTY_SEMISEQUENT.
	    insertFirst(new ConstrainedFormula(add)).semisequent();
	Sequent succSeq = Sequent.createSuccSequent(semSeq);
	TacletGoalTemplate tgt =
	    new AntecSuccTacletGoalTemplate(Sequent.EMPTY_SEQUENT,
					    SLListOfTaclet.EMPTY_LIST, 
					    SLListOfRulevariable.EMPTY_LIST,
					    succSeq, 
					    SetAsListOfSchemaVariable.EMPTY_SET);
	if (precond == Boolean.TRUE){
	    tgt.setName("Precondition");
	}
	else{
	    tgt.setName("Postcondition");
	}
	return tgt;
    }

}
